#include <stdio.h>
#include <inttypes.h>
#include <cuda.h>

#define DCTELEM int16_t

__device__ __constant__ float transform[64]={
    0.353553,  0.353553,  0.353553,  0.353553,  0.353553,  0.353553,  0.353553,  0.353553,
    0.490393,  0.415735,  0.277785,  0.097545, -0.097545, -0.277785, -0.415735, -0.490393,
    0.461940,  0.191342, -0.191342, -0.461940, -0.461940, -0.191342,  0.191342,  0.461940,
    0.415735, -0.097545, -0.490393, -0.277785,  0.277785,  0.490393,  0.097545, -0.415735,
    0.353553, -0.353553, -0.353553,  0.353553,  0.353553, -0.353553, -0.353553,  0.353553,
    0.277785, -0.490393,  0.097545,  0.415735, -0.415735, -0.097545,  0.490393, -0.277785,
    0.191342, -0.461940,  0.461940, -0.191342, -0.191342,  0.461940, -0.461940,  0.191342,
    0.097545, -0.277785,  0.415735, -0.490393,  0.490393, -0.415735,  0.277785, -0.097545
};

__device__ DCTELEM block_d[64];
__device__ float temp_d[64];

__global__ static void mat_idct()
{
    __shared__ float temp[64];
    const uint16_t x = threadIdx.x, y = threadIdx.y;
    uint16_t k;
    float elem = 0.0f;

    for (k = 0; k < 8; k++)
        elem += transform[k*8+x] * block_d[y*8+k];
    temp[y*8+x] = elem;

    __syncthreads();

    elem = 0.0f;
    for (k = 0; k < 8; k++)
        elem += transform[k*8+y] * temp[k*8+x];
    block_d[y*8+x] = lrintf(elem);
}

__global__ static void mat_idct_row()
{
    const uint16_t x = threadIdx.x, y = threadIdx.y;
    uint16_t k;
    float elem = 0.0f;

    for (k = 0; k < 8; k++)
        elem += transform[k*8+x] * block_d[y*8+k];
    temp_d[y*8+x] = elem;
}

__global__ static void mat_idct_col()
{
    const uint16_t x = threadIdx.x, y = threadIdx.y;
    uint16_t k;
    float elem = 0.0f;

    for (k = 0; k < 8; k++)
        elem += transform[k*8+y] * temp_d[k*8+x];
    block_d[y*8+x] = lrintf(elem);
}

void print_mat(DCTELEM block[64])
{
    int x, y;
    for (y = 0; y < 8; y++) {
        printf("\t");
        for (x = 0; x < 8; x++)
            printf("%4hd ", block[y*8+x]);
        printf("\n");
    }
}

void ff_mat_idct_cuda(DCTELEM block[64])
{
    DCTELEM block2[64];

    cudaMemcpyToSymbol(block_d, block, 64*sizeof(DCTELEM));
    mat_idct_row<<< dim3(8,8), dim3(8,8) >>>();
    mat_idct_col<<< dim3(8,8), dim3(8,8) >>>();
    cudaMemcpyFromSymbol(block2, block_d, 64*sizeof(DCTELEM));

    cudaMemcpyToSymbol(block_d, block, 64*sizeof(DCTELEM));
    mat_idct<<< dim3(8,8), dim3(8,8) >>>();
    cudaMemcpyFromSymbol(block, block_d, 64*sizeof(DCTELEM));

    printf("2 kernels, no __threadsync():\n");
    print_mat(block2);
    printf("1 kernel, using __threadsync():\n");
    print_mat(block);
}

int main()
{
    DCTELEM block[64] = {0};
    block[9] = -181;
    block[31] = 39;

    ff_mat_idct_cuda(block);
}
